#!/usr/bin/env python3
from chattymarkov import ChattyMarkov

markov = ChattyMarkov("memory://")
markov.learn("My favorite animal is the crocodile")
markov.learn("The word animal is six letters long")
print(markov.generate())



"""Chattymarkov database submodule.

This submodule gathers all the supported database formats.
"""
import six
from .databases import JSONFileDatabase, MemoryDatabase, RedisDatabase


class ChattymarkovDatabaseError(Exception):
    """Base exception class for chattymarkov.database related errors."""


class UnknownDatabasePrefixError(ChattymarkovDatabaseError):
    """Exception class for unknown database prefixes errors."""


class InvalidConnectionStringError(ChattymarkovDatabaseError):
    """Exception class for invalid connection string error."""


_DATABASE_PREFIXES = {}


def database(prefix):
    """Wrap a function responsible for building a database."""

    def wrapper(func):
        """Register `func` in the global `_DATABASE_PREFIXES` hash."""
        _DATABASE_PREFIXES[prefix] = func
        return func
    return wrapper


def get_database_builder(prefix):
    """Get the function associated to `prefix` to instanciate a database.

    This function is a simple interface around the `_DATABASE_PREFIXES` hash.

    Args:
        prefix (str): the prefix's database function.

    Raises:
        UnknownDatabasePrefixError: the prefix is not recognized.

    Returns:
        function: the function assiociated to the `prefix`.
    """
    if prefix not in _DATABASE_PREFIXES:
        raise UnknownDatabasePrefixError(
            "Database prefix '{}' is unknown.".format(prefix))
    return _DATABASE_PREFIXES[prefix]


def _get_connection_params(resource):
    """Extract connection and params from `resource`."""
    args = resource.split(';')
    if len(args) > 1:
        return args[0], args[1:]
    else:
        return args[0], []


@database('redis')
def build_redis_database(resource):
    """Build a `RedisDatabase` instance to communicate with a redis server.

    Args:
        resource (str): a string that represents connection information.

    Returns:
        RedisDatabase: instance to communicate with the redis server.
    """
    whitelist = {'db'}
    extra_params = {}

    connection, params = _get_connection_params(resource)

    # Parse additional parameters, if any
    if len(params) > 0:
        for param in params:
            key, equal, value = param.partition('=')
            if key in whitelist:
                extra_params[key] = value

    if connection.startswith('/'):
        # UNIX socket connection
        return RedisDatabase(unix_socket_path=connection,
                             **extra_params)
    else:
        # TCP socket connection
        host, colon, port = connection.partition(':')

        if six.PY2:
            port = unicode(port)  # noqa

        if host != '' and colon == ':' and port.isnumeric():
            return RedisDatabase(host=host, port=int(port),
                                 **extra_params)


@database('memory')
def build_memory_database(resource):
    """Build a `MemoryDatabase` instance.

    Args:
        resource (str): path to the memory location. It has actually no sense
            at that time. Should be "memory://" anyway.

    Returns:
        MemoryDatabase: an instance of MemoryDatabase that handles a
            connection to the desired database.
    """
    return MemoryDatabase()


@database('json')
def build_json_database(resource):
    """Build a `JSONFileDatabase` instance.

    Args:
        resource (str): path to the JSON file representing the database. If
            the file is not empty, it will be loaded. In every cases, upon
            instance destruction, the database will be stored in the specified
            file.

    Returns:
        JSONFileDatabase: an instance of JSONFileDatabase that handles a
            connection to the desired database.
    """
    return JSONFileDatabase(resource)


def build_database_connection(connect_string):
    """Build a database connection based on `connect_string`.

    Args:
        connect_string (str): connection string for the database connection.

    Raises:
        InvalidConnectionStringError: raised when the `connect_string` is
            invalid.
        UnknownDatabasePrefixError: raised when the database prefix is
            unknown.
    Returns:
        AbstractDatabase: an instance of AbstractDatabase that handle a
            connection to the desired database.
    """
    prefix, colon_slash_slash, resource = connect_string.partition('://')
    if colon_slash_slash != '':
        builder = get_database_builder(prefix)
        return builder(resource)
    else:
        raise InvalidConnectionStringError(
            "Invalid connection string '{}'. Must be of the form "
            "prefix://[resource[;param1=value1;param2=value2...]]".format(
                prefix))



import unittest

from chattymarkov import database


class TestDatabase(unittest.TestCase):
    def test_build_database_connection(self):
        """Test `build_database_connection` of `chattymarkov.database`."""
        redis_socket_connection = database.build_database_connection(
            "redis:///path/to/redis/unix_socket.sock;db=1")
        self.assertEquals(redis_socket_connection.db, 1)
        self.assertEquals(redis_socket_connection.unix_socket_path,
                          "/path/to/redis/unix_socket.sock")

        redis_socket_connection = database.build_database_connection(
            "redis://localhost:12345;db=3")
        self.assertEquals(redis_socket_connection.db, 3)
        self.assertEquals(redis_socket_connection.host, "localhost")
        self.assertEquals(redis_socket_connection.port, 12345)



import json
import os
import tempfile
import unittest

import six

from chattymarkov import ChattyMarkov
from chattymarkov.database import (JSONFileDatabase, MemoryDatabase,
                                   RedisDatabase)


class TestChattyMarkov(unittest.TestCase):
    def setUp(self):
        empty_memory = []
        self.json_file = tempfile.NamedTemporaryFile(delete=False)
        if six.PY3:
            self.json_file.write(json.dumps(empty_memory).encode())
        else:
            self.json_file.write(json.dumps(empty_memory))
        self.json_file.close()

    def tearDown(self):
        os.unlink(self.json_file.name)

    def test_init(self):
        """Test the __init__ method of ChattyMarkov."""
        c = ChattyMarkov("memory://")
        self.assertTrue(isinstance(c.db, MemoryDatabase))

        c = ChattyMarkov("json://{}".format(self.json_file.name))
        self.assertTrue(isinstance(c.db, JSONFileDatabase))

        c = ChattyMarkov("redis:///path/to/unix_socket_file.sock")
        self.assertTrue(isinstance(c.db, RedisDatabase))



"""Database interfaces for chattimarkov."""
import atexit
import json
import os.path
import random

import redis
import six

from .base import AbstractDatabase


class RedisDatabase(AbstractDatabase):
    def __init__(self, host="localhost", port=6739, db=0,
                 unix_socket_path=None):

        self._db = int(db)

        if unix_socket_path is not None:
            self._unix_socket_path = unix_socket_path
            self.handle = redis.StrictRedis(unix_socket_path=unix_socket_path,
                                            db=db)
        else:
            self._host = host
            self._port = int(port)
            self.handle = redis.StrictRedis(host=host, port=port, db=db)

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def db(self):
        return self._db

    @property
    def unix_socket_path(self):
        return self._unix_socket_path

    def add(self, key, element):
        return self.handle.sadd(key, element.encode()) > 0

    def random(self, key):
        element = self.handle.srandmember(key)
        if element is not None:
            return element.decode()

    def get(self, key):
        element = self.handle.get(key)
        if element is not None:
            return element.decode()

    def set(self, key, value):
        self.handle.set(key, value)


class MemoryDatabase(AbstractDatabase):
    """Memory database class for chattimarkov.

    This is just a volatile, in-memory database which is built either from a
    pre-existing dictionary or from scratch. Upon object destruction, the
    database is not saved.
    """

    def __init__(self, db=None, *args, **kwargs):
        if db is None:
            self.db = {}
        else:
            self.db = db

    def add(self, key, element):
        if key not in self.db:
            self.db[key] = []
        if type(self.db[key]) is not list:
            return False

        if element not in self.db[key]:
            self.db[key].append(element)

    def random(self, key):
        if key not in self.db:
            return None
        if type(self.db[key]) is not list:
            return None
        return random.choice(self.db[key])

    def get(self, key, default=None):
        return self.db.get(key, default)

    def set(self, key, value):
        self.db[key] = value


class JSONFileDatabase(MemoryDatabase):
    """JSON database class for Chattimarkov.

    JSON files are actually not really suitable for storing a key-value
    database. Loading and storing can actually be slow, mostly if the
    chattimarkov instance is learning a lot.
    """

    def __init__(self, filepath, *args, **kwargs):
        self.filepath = filepath
        if os.path.exists(self.filepath):
            db = json.load(open(filepath))
        else:
            db = None

        if six.PY2:
            MemoryDatabase.__init__(self, db, *args, **kwargs)
        else:
            super().__init__(db, *args, **kwargs)

        atexit.register(self.cleanup)

    def cleanup(self):
        with open(self.filepath, "w") as stream:
            json.dump(self.db, stream)



from chattymarkov.database import MemoryDatabase


class TestMemoryDatabase:
    def setup_method(self, method):
        self.db = MemoryDatabase()

    def test_add(self):
        """Test the `add` method from the MemoryDatabase class."""
        self.db.add('foo', 'bar')
        assert 'foo' in self.db.db
        assert 'bar' in self.db.db['foo']
